#!/usr/bin/python3
ver="2025-01-05"
############################################################
# REFDOI Ver. 2025-01-05
# (Utility to obtain DOIs for references in EXFOR file)
#
# Naohiko Otuka (IAEA Nuclear Data Section)
############################################################
import datetime
import json
import os
import re
import argparse

import x4_x4toj4
import x4_refbib

def main(file_inp,file_dict,file_out,email,force0):

# Input parameters for X4TOJ4 and REFDOI
  force=force0

# Input parameters for X4TOJ4
  time=datetime.datetime.now()
  file_tmp= time.strftime('%Y%m%d%H%M%S%f')
  chkrid  = False  # record identificaiton not checked in X4TOJ4
  keykeep = "REFERENCE REL-REF MONIT-REF" # keep reference keywords only
  keepflg = False  # ignore flags at col.80 in X4TOJ4
  outstr  = True   # print real numbers as strings in X4TOJ4
  delpoin = True   # delete pointer in the output in POIPOI
  keep001 = True   # keep common subentry in POIPOI

# Input parameters for REFDOI
  x4_doi=None
  format="doi"
  strip=False
  fauthor=""

# Conversion to J4 format by X4TOJ4
  x4_x4toj4.main(file_inp,file_dict,file_tmp,force,chkrid,keykeep,keepflg,outstr)
  x4_json_full=read_x4json(file_tmp)
  os.remove(file_tmp)

# Extract reference codes
  ansans=[]
  keywords=[]
  codes=[]

  for nentry, entry in enumerate(x4_json_full["entries"]):
    x4_json=dict()
    x4_json=x4_json_full["entries"][nentry]

    for nsubent, subentry in enumerate(x4_json["subentries"]):

      if "SUBENT" in subentry:
        ansan=subentry["SUBENT"]["N1"]
        ansan=ansan[0:5]+"."+ansan[5:8]

        if "REFERENCE" in subentry:
          for reference in subentry["REFERENCE"]:
            if "coded_information" in reference:
              for code_unit in reference["coded_information"]["code_unit"]:
                ansans.append(ansan)
                keywords.append("REFERENCE") 
                codes.append(code_unit["unit"])

        if "REL-REF" in subentry:
           for relref in subentry["REL-REF"]:
            if "coded_information" in relref:
              ansans.append(ansan)
              keywords.append("REL-REF") 
              codes.append(relref["coded_information"]["reference"]["code"]) 

        if "MONIT-REF" in subentry:
          for monitref in subentry["MONIT-REF"]:
            if "coded_information" in monitref:
              ansans.append(ansan)
              keywords.append("MONIT-REF") 
              codes.append(monitref["coded_information"]["reference"]["code"])
 
# Check with REFDOI
  f=open(file_out, "w")
  errors=[]
  for ansan, keyword, code in zip(ansans,keywords,codes):
    x4_refbib.main(code,x4_doi,file_dict,file_tmp,format,email,force,strip,fauthor)
    g=open(file_tmp, "r")
    lines=g.readlines()
    g.close()
    os.remove(file_tmp)

    keyword="%-10s" % keyword
    code="%-30s" % code

    if lines[0][0:2]=="**":
      if re.compile("DOI").search(lines[0]):
        char="** Suspiciuos reference code"
        errors.append(ansan+" "+keyword+" "+code)
      else:
        char="** DOI presence not checked"

    else:
      char=lines[0]
      
    f.write(ansan+" "+keyword+" "+code+" "+char+"\n")

  print()
  print("-----------------------------------------")
  if len(errors)==0:
    print ("Suspicious reference not code detected")
  else:
    print (str(len(errors))+" suspicious reference code(s) detected")
    print()
    for error in errors:
      print(error)
  print("-----------------------------------------")

  print("REFDOI: Processing terminated normally.")


def read_x4json(file_x4):
  f=open(file_x4)
  try:
    x4_json=json.load(f)
  except json.JSONDecodeError:
    msg=file_x4+" is not in JSON format."
    print_error_fatal(msg,"")

  if x4_json["name"][0:18]!="J4 - EXFOR in JSON":
    msg=file_x4+" is not an EXFOR in JSON."
    print_error_fatal(msg,"")

  f.close()
  return x4_json


def get_args(ver):

  parser=argparse.ArgumentParser(\
   usage="Check presence of references in EXFOR file",\
   epilog="example: x4_refdoi.py -i exfor.txt -d dict_9131.json -o x4_refdoi_out.txt -e email@address.com")
  parser.add_argument("-v", "--version",\
   action="version", version=ver)
  parser.add_argument("-f", "--force",\
   help="never prompt", action="store_true")
  parser.add_argument("-i", "--file_inp",\
   help="input EXFOR file")
  parser.add_argument("-d", "--file_dict",\
   help="input JSON dictionary")
  parser.add_argument("-o", "--file_out",\
   help="output DOI file")
  parser.add_argument("-e", "--email",\
   help="your email address")

  args=parser.parse_args()
  return args


def get_input(args):
  time=datetime.datetime.now()
  date=time.strftime("%Y-%m-%d")
  print("REFDOI (Ver-"+ver+") run on "+date)
  print("-----------------------------------------")

  force0=args.force

  file_inp=args.file_inp
  if file_inp is None:
    file_inp=input("input EXFOR file [exfor.txt] --------> ")
    if file_inp=="":
      file_inp="exfor.txt"
  if not os.path.exists(file_inp):
    print(" ** File "+file_inp+" does not exist.")
  while not os.path.exists(file_inp):
    file_inp=input("input EXFOR file [exfor.txt] -----------> ")
    if file_inp=="":
      file_inp="exfor.txt"
    if not os.path.exists(file_inp):
      print(" ** File "+file_inp+" does not exist.")

  file_dict=args.file_dict
  if file_dict is None:
    file_dict=input("JSON Dictionary [dict_9131.json] ----> ")
    if file_dict=="":
      file_dict="dict_9131.json"
  if not os.path.exists(file_dict):
    print(" ** File "+file_dict+" does not exist.")
  while not os.path.exists(file_dict):
    file_dict=input("JSON Dictionary [dict_9131.json] -> ")
    if file_dict=="":
      file_dict="dict_9131.json"
    if not os.path.exists(file_dict):
      print(" ** File "+file_dict+" does not exist.")

  file_out=args.file_out
  if file_out is None:
    file_out=input("Output DOI file [x4_refdoi_out.txt] -> ")
  if file_out=="":
    file_out="x4_refdoi_out.txt"
  if os.path.isfile(file_out):
    msg="File '"+file_out+"' exists and must be overwritten."
    print_error(msg,"",force0)

  email=args.email
  if email is None:
    email=input("your email address ------------------> ")
  if not re.compile("^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}$").search(email):
    print(" ** Input a correct email address.")
  while not re.compile("^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}$").search(email):
    email=input("your email address ------------------> ")
    if not re.compile("^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}$").search(email):
      print(" ** Input a correct email address.")

  return file_inp,file_dict,file_out,email,force0


def print_error(msg,line,force):
  print("** "+msg)
  print(line)

  if force:
    answer="Y"
  else:
    answer=""

  while answer!="Y" and answer!="N":
    answer=input("Continue? [Y] --> ")
    if answer=="":
        answer="Y"
    if answer!="Y" and answer!="N":
      print(" ** Answer must be Y (Yes) or N (No).")
  if answer=="N":
    print("program terminated")
    exit()


if __name__ == "__main__":
  args=get_args(ver)
  (file_inp,file_dict,file_out,email,force0)=get_input(args)
  main(file_inp,file_dict,file_out,email,force0)
  exit()
